/*----------------------------------------------------------------------------------------------

Gewitter-Monitor

Stephan Laage-Witt - Oktober 2014
Version 1.09

---------------------------------------------------------------------------------------------------*/

#define F_CPU 8000000					// 8 MHz Quarz

#define IRQ_FREQ 500					// Anzahl Interrupts pro Sekunde
#define STRIKE_LED_TIME IRQ_FREQ / 2 	// Blinkzeit der LED: eine halbe Sekunde 
#define STEEPNESS_THRESHOLD 150			// Schwellwert fr Flankendetektor
#define ALARM_TIME 300					// Zeit, nach der die Alarm-Level zurckgesetzt wird (sec)
#define BACKLIGHT_TIME 30				// LCD-Backlight-Zeit nach Tastendruck (sec)	
#define ALARM_1_THRESHOLD 250			// Schwellwert fr Alarm-Level 1
#define ALARM_2_THRESHOLD 500			// Schwellwert fr Alarm-Level 2

// Pin-Belegung
#define PIN_STRIKE_LED_NS 4
#define PIN_STRIKE_LED_OW 5
#define PIN_ALARM_LED 3
#define PIN_LCD_BACKLIGHT 2
#define PIN_LCD_BACKLIGHT_SWITCH 0

#include "u8g.h"
#include <avr/interrupt.h>
#include <avr/io.h>
#include <avr/sleep.h>

// Globale Variablen, die von der INterruptroutine verndert werden
volatile uint8_t to_do_flag;			// zur Kommunikation zwischen Interrupt und Hauptprogramm
volatile uint16_t max_ns = 0,			// Peak-Werte der Blitze und Summe pro Minute 
				  max_sum_ns = 0, 
				  max_ow = 0, 
				  max_sum_ow = 0;
volatile uint16_t backlight_counter = ALARM_TIME;		// Verbleibende Zeit der LCD-Hintergrundbeleuchtung
volatile uint8_t sec = 0;				// Sekundenzhler

// Globale Variablen ohne Interrupt-Beteiligung
u8g_t u8g;								// Graphik-Handle, wird von U8glib verwaltet
uint8_t strike_array_ns[100],			// Umgerechnete Summe pro Minute fr Display 
		strike_array_ow[100];
uint8_t strike_array_pointer = 0;				// Pointer fr strike_arrays
char time_string[3] = "00";				// Display: Sekundenzhler
char string_max_ns[6],					// Display: letzter registrierter Blitz 
	 string_max_sum_ns[6],				// Display: Summe fr die laufende Minute
	 string_max_ow[6],					// Display: letzter registrierter Blitz 
	 string_max_sum_ow[6],				// Display: Summe fr die laufende Minute
	 string_b[6] = "  0.0";				// Display: Steigung der linearen Regression


/* Interrupt Service Handler fr 16 Bit Timer 1 --------------------------------------------------
   Fragt ADC ab, abwechselnd fr die beiden Kanle. Sucht ansteigende Flanke und Maximum. 
   Schaltet Strike-LEDs.
*/

ISR(TIMER1_COMPA_vect) {
	static uint16_t counter = 0, old_value_ns, old_value_ow;
	static uint16_t strike_LED_counter_ns = 0, strike_LED_counter_ow = 0; 
	static uint8_t channel = 0, mode_ns, mode_ow; 
	uint16_t value;

	if (++counter == IRQ_FREQ) {		// Einmal pro Sekunde das Hauptprogramm aktivieren
		to_do_flag |= 1; 
		counter = 0;
		if (++sec == 60) sec = 0;
	};
	
	if ((PINB & (1<<PIN_LCD_BACKLIGHT_SWITCH)) == 0) {		// Taste fr LCD-Beleuchtung gedrckt?
		PORTC &= ~(1 << PIN_LCD_BACKLIGHT);					// LCD-Beleuchtung einschalten
		backlight_counter = BACKLIGHT_TIME;
	}
	
	value = ADCW;						// ADC auslesen
	if (value > 1000) value = 1000;		// Werte werden auf 1000 begrenzt

	if (channel == 0) {					// Nord-Sd abfragen
		switch (mode_ns) {
			case 0:						// Ansteigende Flanke suchen 
				if (value > (old_value_ns + STEEPNESS_THRESHOLD)) {
					mode_ns = 1;
				};			
				break;
			case 1:						// Maximum suchen 
				if (value < old_value_ns) {			// Maximum erreicht?
					max_ns = old_value_ns / 10;		// Maximum auf 0...100 reduzieren
					max_sum_ns += max_ns;
					mode_ns = 0;
					PORTC &= ~(1 << PIN_STRIKE_LED_NS);	// Strike-LED einschalten
					strike_LED_counter_ns = STRIKE_LED_TIME;				
				};				
				break;
		}
	} else {							// Ost-West abfragen
		switch (mode_ow) {
			case 0:						// Ansteigende Flanke suchen
				if (value > (old_value_ow + STEEPNESS_THRESHOLD)) {
					mode_ow = 1;
				};			
				break;
			case 1:						// Maximum suchen
				if (value < old_value_ow) {			// Maximum erreicht?
					max_ow = old_value_ow / 10;		// Maximum auf 0...100 reduzieren
					max_sum_ow += max_ow;
					mode_ow = 0;
					PORTC &= ~(1 << PIN_STRIKE_LED_OW);	// Strike-LED einschalten
					strike_LED_counter_ow = STRIKE_LED_TIME;				
				};				
				break;
		};			
	};

	// Strike-LEDs checken und gegebenenfalls abschalten
	if (strike_LED_counter_ns > 0) --strike_LED_counter_ns;
	else PORTC |= (1 << PIN_STRIKE_LED_NS);				// Strike-LED 1 ausschalten	
	if (strike_LED_counter_ow > 0) --strike_LED_counter_ow;
	else PORTC |= (1 << PIN_STRIKE_LED_OW);				// Strike-LED 2 ausschalten

	// ADC neu starten
	ADMUX  = (1<<REFS1) | (0<<ADLAR) | channel;			// Kanal setzen			
	ADCSRA |= (1<<ADSC);								// ADC starten 
	
	// Vorherigen Wert sichern und ADC-Kanal setzen
	if (channel == 0) {
		old_value_ns = value;
		channel = 1;
	} else { 
		old_value_ow = value;
		channel = 0;
	};	
}	


/* Draw-Routine fr u8g-Graphik ----------------------------------------------------------- */

void draw(void)
{
	uint8_t i;
	uint8_t strike_value_ns, strike_value_ow;		// Lokale Kopie der Strike-Values

	// Aktuelle Daten auf der rechten Display-Seite ausgeben
	u8g_SetFont(&u8g, u8g_font_5x8);
	u8g_DrawStr(&u8g, 119, 6, time_string);
	u8g_DrawLine(&u8g, 119, 8, 127, 8);
	u8g_DrawStr(&u8g, 114, 18, string_max_ow);
	u8g_DrawStr(&u8g, 109, 26, string_max_sum_ow);
	u8g_DrawLine(&u8g, 114, 28, 127, 28);
	u8g_DrawStr(&u8g, 114, 38, string_max_ns);
	u8g_DrawStr(&u8g, 109, 46, string_max_sum_ns);
	u8g_DrawLine(&u8g, 109, 48, 127, 48);
	u8g_DrawStr(&u8g, 104, 58, string_b);
	
	// Graphik ausgeben
	for (i = 0; i < 100; ++i) {
		strike_value_ns = strike_array_ns[strike_array_pointer];
		strike_value_ow = strike_array_ow[strike_array_pointer];
		
		// Falls Alarm-Level-2 gesetzt ist, Werte halbieren fr erweiterte Skala (0 ... 1200)
		if ((PORTC & (1<<PIN_ALARM_LED)) == 0) {
			strike_value_ns /= 2;
			strike_value_ow /= 2;
		};
		
		// Werte auf Displaygre begrenzen (0 ... 63)
		if ((strike_value_ns + strike_value_ow) > 63) {
			strike_value_ns = 63;
			strike_value_ow = 0;
		};
		
		// Senkrechte Pixelreihe zeichnen
		u8g_DrawLine(&u8g, i, 63, i, 63 - strike_value_ns - strike_value_ow);
		if (strike_value_ns > 0) {
			u8g_SetColorIndex(&u8g, 0);
			u8g_DrawPixel(&u8g, i, 63 - strike_value_ns);
			u8g_SetColorIndex(&u8g, 1);
		};			
		
		// Pointer heraufzhlen
		if (++strike_array_pointer == 100)
			strike_array_pointer = 0;
	};

	// Y-Skalenpunkte und Beschriftung zeichnen 
	u8g_SetFont(&u8g, u8g_font_4x6);
	if ((PORTC & (1<<PIN_ALARM_LED)) == 0) {
		for (i = 1; i < 13; ++i)
			u8g_DrawPixel(&u8g, 100, 63 - i * 5);
		u8g_DrawStr(&u8g, 102, 6, "1200");	
		u8g_DrawStr(&u8g, 102, 36, "600");	
	} else {
		for (i = 1; i < 7; ++i)
			u8g_DrawPixel(&u8g, 100, 63 - i * 10);
		u8g_DrawStr(&u8g, 102, 6, "600");	
		u8g_DrawStr(&u8g, 102, 36, "300");	
	};	
	
	// X-Skalenpunkte alle 10 Minuten zeichnen
	for (i = 0; i < 100; i += 10) {
		u8g_DrawPixel(&u8g, i, 0);
	};
	
}


/* lcd_dec_unit32 --------------------------------------------------------------------------------*/
/* Gibt einen 32-Bit-Wert als Dezimalzahl aus.													  */
/* positions -> Anzahl der Stellen (2 ... 7)													  */
/* dec_point -> Anzahl der Nachkommastellen	(0 ... positions)									  */

void lcd_dec_uint32(uint32_t value, char string_array[], uint8_t positions, const uint8_t dec_point)
{
	uint8_t count, blank_flag = 0, pos = 0;
	uint32_t denominator = 10;

	for (count = 2; count < positions; ++count)
		denominator *= 10;
	 	
	while (positions > 0) {
		count = value / denominator;
		value -= count * denominator;
		if (positions == dec_point + 1) blank_flag = 1;
		if (positions == dec_point) {
			string_array[pos] = '.';
			++pos;
		};			
		if ((count == 0) && (blank_flag == 0)) {
			string_array[pos] =' ';
			++pos;
		} else {
			string_array[pos] = '0' + count;
			++pos;
			blank_flag = 1;
		};			
		--positions;
		denominator /= 10;
	};					
	string_array[pos] = 0;
}


/* Calculate_B ------------------------------------------------------------------------------------*/
/* Simplifizierte lineare Regression ber die letzten 30 Minuten                                   */
/* Gibt den berechneten Steigungskoeffizienten B zurck                                            */

int16_t calculate_b (uint8_t pointer) {
	int8_t i, value;
	int32_t sum_value, sum_valuexi;
	int16_t result;
	
	sum_value = 0;
	sum_valuexi = 0;
	for (i = 0; i < 30; ++i) {
		value = strike_array_ns[pointer] + strike_array_ow[pointer];
		sum_value += value;
		sum_valuexi += (value * i);
		if (pointer == 0) pointer = 99;
		else --pointer;
	};
	result = (87 * sum_value - 6 * sum_valuexi) / 135;
	return(result);
};


/* Hauptprogramm -----------------------------------------------------------------------------------*/

int main(void) {
	uint8_t i;
	int16_t b;
	uint16_t alarm_counter = 0;
	
	// Input/Output-Register setzen
	DDRB = (0 << PIN_LCD_BACKLIGHT_SWITCH);				// Pin ist Input fr LCD-Beleuchtungs-Taster
	PORTB = (1 << PIN_LCD_BACKLIGHT_SWITCH);			// Pull-Up Widerstand einschalten
	DDRC = 0b11111100;									// ADC-Eingnge sind Input
	PORTC = 0b11111100;									// Alle Funktionen ausschalten
	
	// Sleep Mode setzen: Timer und Interrupts laufen weiter
	set_sleep_mode(SLEEP_MODE_IDLE);
	
	// ADC initialisieren 
	ADMUX  = (1<<REFS0) | (0<<ADLAR);								// 10 bit, right adjusted. Interne Referenz verwenden				
	ADCSRA = (1<<ADEN) | (1<<ADPS2) | (1<<ADPS1) | (1<<ADPS0);		// scaling factor: 128 -> 125 kHz (page 257)  
	ADMUX  = (1<<REFS1) | (0<<ADLAR) | 0;							// Kanal setzen			
	ADCSRA |= (1<<ADSC);											// ADC starten
	
 	// U8GLIB initialisieren 
	u8g_InitSPI(&u8g, &u8g_dev_st7920_128x64_4x_hw_spi, PN(1, 5), PN(1, 3), PN(1, 2), PN(1, 1), U8G_PIN_NONE);

	// Strike-Arrays initialisieren
	for (i = 0; i < 100; ++i) { 
		strike_array_ns[i] = 0;
		strike_array_ow[i] = 0;
	};	

	// 16-Bit Timer1 starten, 8000000 / 64 / 250 => 500 Hz 
	OCR1A = 250-1;											  
	TCCR1B |= (1<<WGM12)| (0<<CS02) | (1<<CS01) |(1<<CS00); 		// CPU-Frequenz / 64, CTC1 = Clear Timer on Compare
	TIMSK1 |= (1 << OCIE1A); 	
	sei();															// Interrupt aktivieren - los geht's

	while (1) {

		if (to_do_flag) {				// einmal pro Sekunde ...
			to_do_flag = 0;				// To-Do-Flag lschen

			time_string[0] = (sec / 10) + '0';				// Sekundenzhler als String fr die Ausgabe vorbereiten
			time_string[1] = (sec % 10) + '0';
			
			if (sec == 0) {				// einmal pro Minute ...
				strike_array_ns[strike_array_pointer] = max_sum_ns / 10;		// Summen pro Minute auf Graphik-Skala umrechnen ...
				strike_array_ow[strike_array_pointer] = max_sum_ow / 10;		// ... und in das Strike-Array bernehmen
				max_sum_ns = 0;													
				max_sum_ow = 0;	
				if (++strike_array_pointer == 100)					// Strike-Array-Pointer heraufzhlen
					strike_array_pointer = 0;

				// Lineare Regression berechnen und String-Ausgabe mit Vorzeichen als String vorbereiten
				b = calculate_b(strike_array_pointer);
				if (b < 0) {
					b = -b;
					string_b[0] = '-';
				} else if (b > 0) {
					string_b[0] = '+';
				} else {
					string_b[0] = ' ';
				};	
				lcd_dec_uint32(b, string_b + 1, 3, 1);
			};

			// LCD-Backlight und Alarm-LED setzen
			if ((max_sum_ns + max_sum_ow) >= ALARM_1_THRESHOLD) {
				PORTC &= ~(1 << PIN_LCD_BACKLIGHT);				// LCD-Hintergrundbeleuchtung einschalten
				backlight_counter = ALARM_TIME;
			};
			if ((max_sum_ns + max_sum_ow) > ALARM_2_THRESHOLD) {
				PORTC &= ~(1 << PIN_ALARM_LED);
				alarm_counter = ALARM_TIME;
			};

			if (!(PORTC & (1 << PIN_LCD_BACKLIGHT))) {			// Display-Ausgabe nur, wenn die LCD-Beleuchtung aktiv ist

				// Strike-Werte als String fr die Ausgabe vorbereiten
				lcd_dec_uint32(max_ns, string_max_ns, 3, 0);
				lcd_dec_uint32(max_sum_ns, string_max_sum_ns, 4, 0);
				lcd_dec_uint32(max_ow, string_max_ow, 3, 0);
				lcd_dec_uint32(max_sum_ow, string_max_sum_ow, 4, 0);
				
				// Graphik-Ausgabe
				u8g_FirstPage(&u8g);					
				do {
					draw();
				} while (u8g_NextPage(&u8g));
			
			};

			// LCD-Backlight und Alarm-LED checken und gegebenenfalls ausschalten
			if (backlight_counter > 0) --backlight_counter;
			else PORTC |= (1 << PIN_LCD_BACKLIGHT);					
			if (alarm_counter > 0) --alarm_counter;
			else PORTC |= (1 << PIN_ALARM_LED);
			
		} else {
			sleep_mode();
		};		
	};	 
}